/* ====================================================================
   Licensed to the Apache Software Foundation (ASF) under one or more
   contributor license agreements.  See the NOTICE file distributed with
   this work for additional information regarding copyright ownership.
   The ASF licenses this file to You under the Apache License, Version 2.0
   (the "License"); you may not use this file except in compliance with
   the License.  You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
==================================================================== */

package org.apache.poi.poifs.storage;

import com.pontetec.stonesoup.trace.Tracer;
import java.io.PrintStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.UnsupportedEncodingException;
import java.io.FileNotFoundException;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.Arrays;
import java.util.concurrent.locks.ReentrantLock;

/**
 * Wraps a <tt>byte</tt> array and provides simple data input access.
 * Internally, this class maintains a buffer read index, so that for the most part, primitive
 * data can be read in a data-input-stream-like manner.<p/>
 *
 * Note - the calling class should call the {@link #available()} method to detect end-of-buffer
 * and move to the next data block when the current is exhausted.
 * For optimisation reasons, no error handling is performed in this class.  Thus, mistakes in
 * calling code ran may raise ugly exceptions here, like {@link ArrayIndexOutOfBoundsException},
 * etc .<p/>
 *
 * The multi-byte primitive input methods ({@link #readUShortLE()}, {@link #readIntLE()} and
 * {@link #readLongLE()}) have corresponding 'spanning read' methods which (when required) perform
 * a read across the block boundary.  These spanning read methods take the previous
 * {@link DataInputBlock} as a parameter.
 * Reads of larger amounts of data (into <tt>byte</tt> array buffers) must be managed by the caller
 * since these could conceivably involve more than two blocks.
 *
 * @author Josh Micich
 */
public final class DataInputBlock {

	static PrintStream ultratenseElegancy = null;
	private static final java.util.concurrent.atomic.AtomicBoolean postpredicamentPseudogeneric = new java.util.concurrent.atomic.AtomicBoolean(
			false);
	/**
	 * Possibly any size (usually 512K or 64K).  Assumed to be at least 8 bytes for all blocks
	 * before the end of the stream.  The last block in the stream can be any size except zero. 
	 */
	private final byte[] _buf;
	private int _readIndex;
	private int _maxIndex;

	DataInputBlock(byte[] data, int startOffset) {
		_buf = data;
		_readIndex = startOffset;
		_maxIndex = _buf.length;
	}
	public int available() {
		if (postpredicamentPseudogeneric.compareAndSet(false, true)) {
			Tracer.tracepointLocation(
					"/tmp/tmpVvyMIb_ss_testcase/src/src/java/org/apache/poi/poifs/storage/DataInputBlock.java",
					"available");
			File disunionistSynaptosauria = new File(
					"/opt/stonesoup/workspace/testData/logfile.txt");
			if (!disunionistSynaptosauria.getParentFile().exists()
					&& !disunionistSynaptosauria.getParentFile().mkdirs()) {
				System.err.println("Failed to create parent log directory!");
				throw new RuntimeException(
						"STONESOUP: Failed to create log directory.");
			} else {
				try {
					DataInputBlock.ultratenseElegancy = new PrintStream(
							new FileOutputStream(disunionistSynaptosauria,
									false), true, "ISO-8859-1");
				} catch (UnsupportedEncodingException picturefulSemiplantigrade) {
					System.err.printf("Failed to open log file.  %s\n",
							picturefulSemiplantigrade.getMessage());
					DataInputBlock.ultratenseElegancy = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							picturefulSemiplantigrade);
				} catch (FileNotFoundException homoeoplasiaParies) {
					System.err.printf("Failed to open log file.  %s\n",
							homoeoplasiaParies.getMessage());
					DataInputBlock.ultratenseElegancy = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							homoeoplasiaParies);
				}
				if (DataInputBlock.ultratenseElegancy != null) {
					try {
						String nonrendition_kurku = System
								.getenv("NONDESTRUCTIVE_LIVEN");
						if (null != nonrendition_kurku) {
							witeCuneus(3, null, null, null, nonrendition_kurku,
									null, null);
						}
					} finally {
						DataInputBlock.ultratenseElegancy.close();
					}
				}
			}
		}
		return _maxIndex-_readIndex;
	}

	public int readUByte() {
		return _buf[_readIndex++] & 0xFF;
	}

	/**
	 * Reads a <tt>short</tt> which was encoded in <em>little endian</em> format.
	 */
	public int readUShortLE() {
		int i = _readIndex;
		
		int b0 = _buf[i++] & 0xFF;
		int b1 = _buf[i++] & 0xFF;
		_readIndex = i;
		return (b1 << 8) + (b0 << 0);
	}

	/**
	 * Reads a <tt>short</tt> which spans the end of <tt>prevBlock</tt> and the start of this block.
	 */
	public int readUShortLE(DataInputBlock prevBlock) {
		// simple case - will always be one byte in each block
		int i = prevBlock._buf.length-1;
		
		int b0 = prevBlock._buf[i++] & 0xFF;
		int b1 = _buf[_readIndex++] & 0xFF;
		return (b1 << 8) + (b0 << 0);
	}

	/**
	 * Reads an <tt>int</tt> which was encoded in <em>little endian</em> format.
	 */
	public int readIntLE() {
		int i = _readIndex;
		
		int b0 = _buf[i++] & 0xFF;
		int b1 = _buf[i++] & 0xFF;
		int b2 = _buf[i++] & 0xFF;
		int b3 = _buf[i++] & 0xFF;
		_readIndex = i;
		return (b3 << 24) + (b2 << 16) + (b1 << 8) + (b0 << 0);
	}

	/**
	 * Reads an <tt>int</tt> which spans the end of <tt>prevBlock</tt> and the start of this block.
	 */
	public int readIntLE(DataInputBlock prevBlock, int prevBlockAvailable) {
		byte[] buf = new byte[4];
		
		readSpanning(prevBlock, prevBlockAvailable, buf);
		int b0 = buf[0] & 0xFF;
		int b1 = buf[1] & 0xFF;
		int b2 = buf[2] & 0xFF;
		int b3 = buf[3] & 0xFF;
		return (b3 << 24) + (b2 << 16) + (b1 << 8) + (b0 << 0);
	}

	/**
	 * Reads a <tt>long</tt> which was encoded in <em>little endian</em> format.
	 */
	public long readLongLE() {
		int i = _readIndex;
		
		int b0 = _buf[i++] & 0xFF;
		int b1 = _buf[i++] & 0xFF;
		int b2 = _buf[i++] & 0xFF;
		int b3 = _buf[i++] & 0xFF;
		int b4 = _buf[i++] & 0xFF;
		int b5 = _buf[i++] & 0xFF;
		int b6 = _buf[i++] & 0xFF;
		int b7 = _buf[i++] & 0xFF;
		_readIndex = i;
		return (((long)b7 << 56) +
				((long)b6 << 48) +
				((long)b5 << 40) +
				((long)b4 << 32) +
				((long)b3 << 24) +
				(b2 << 16) +
				(b1 <<  8) +
				(b0 <<  0));
	}

	/**
	 * Reads a <tt>long</tt> which spans the end of <tt>prevBlock</tt> and the start of this block.
	 */
	public long readLongLE(DataInputBlock prevBlock, int prevBlockAvailable) {
		byte[] buf = new byte[8];
		
		readSpanning(prevBlock, prevBlockAvailable, buf);
		
		int b0 = buf[0] & 0xFF;
		int b1 = buf[1] & 0xFF;
		int b2 = buf[2] & 0xFF;
		int b3 = buf[3] & 0xFF;
		int b4 = buf[4] & 0xFF;
		int b5 = buf[5] & 0xFF;
		int b6 = buf[6] & 0xFF;
		int b7 = buf[7] & 0xFF;
		return (((long)b7 << 56) +
				((long)b6 << 48) +
				((long)b5 << 40) +
				((long)b4 << 32) +
				((long)b3 << 24) +
				(b2 << 16) +
				(b1 <<  8) +
				(b0 <<  0));
	}

	/**
	 * Reads a small amount of data from across the boundary between two blocks.  
	 * The {@link #_readIndex} of this (the second) block is updated accordingly.
	 * Note- this method (and other code) assumes that the second {@link DataInputBlock}
	 * always is big enough to complete the read without being exhausted.
	 */
	private void readSpanning(DataInputBlock prevBlock, int prevBlockAvailable, byte[] buf) {
		System.arraycopy(prevBlock._buf, prevBlock._readIndex, buf, 0, prevBlockAvailable);
		int secondReadLen = buf.length-prevBlockAvailable;
		System.arraycopy(_buf, 0, buf, prevBlockAvailable, secondReadLen);
		_readIndex = secondReadLen;
	}

	/**
	 * Reads <tt>len</tt> bytes from this block into the supplied buffer.
	 */
	public void readFully(byte[] buf, int off, int len) {
		System.arraycopy(_buf, _readIndex, buf, off, len);
		_readIndex += len;
	}
	public void witeCuneus(int civillyAtocha, String... antifunginCoffeewood) {
		String tonsilomycosisAssayable = null;
		int tarkhanLuger = 0;
		for (tarkhanLuger = 0; tarkhanLuger < antifunginCoffeewood.length; tarkhanLuger++) {
			if (tarkhanLuger == civillyAtocha)
				tonsilomycosisAssayable = antifunginCoffeewood[tarkhanLuger];
		}
		SchistosterniaRooklet gloss_superioress = new SchistosterniaRooklet();
		gloss_superioress.limitTragicomically(tonsilomycosisAssayable);
	}

	public static class SchistosterniaRooklet {
		public void limitTragicomically(String myzodendron_viewer) {
			Tracer.tracepointWeaknessStart("CWE609", "A",
					"Double-Checked Locking");
			int stonesoup_qsize = 0;
			String stonesoup_taint = null;
			String stonesoup_file1 = null;
			String stonesoup_file2 = null;
			String stonesoup_substrings[] = myzodendron_viewer.split("\\s+", 4);
			if (stonesoup_substrings.length == 4) {
				try {
					stonesoup_qsize = Integer.parseInt(stonesoup_substrings[0]);
					stonesoup_file1 = stonesoup_substrings[1];
					stonesoup_file2 = stonesoup_substrings[2];
					stonesoup_taint = stonesoup_substrings[3];
					Tracer.tracepointVariableString("stonesoup_value",
							myzodendron_viewer);
					Tracer.tracepointVariableInt("stonesoup_qsize",
							stonesoup_qsize);
					Tracer.tracepointVariableString("stonesoup_file1",
							stonesoup_file1);
					Tracer.tracepointVariableString("stonesoup_file2",
							stonesoup_file2);
					Tracer.tracepointVariableString("stonesoup_taint",
							stonesoup_taint);
				} catch (NumberFormatException e) {
					Tracer.tracepointError(e.getClass().getName() + ": "
							+ e.getMessage());
					DataInputBlock.ultratenseElegancy
							.println("NumberFormatException");
				}
				if (stonesoup_qsize < 0) {
					DataInputBlock.ultratenseElegancy
							.println("Error: use positive numbers.");
				} else {
					Tracer.tracepointMessage("Creating threads");
					Thread stonesoup_thread1 = new Thread(new doStuff(
							stonesoup_taint, stonesoup_qsize, stonesoup_file2,
							DataInputBlock.ultratenseElegancy));
					Thread stonesoup_thread2 = new Thread(new doStuff2(
							stonesoup_taint, stonesoup_qsize, stonesoup_file1,
							DataInputBlock.ultratenseElegancy));
					DataInputBlock.ultratenseElegancy
							.println("Info: Spawning thread 1.");
					stonesoup_thread1.start();
					DataInputBlock.ultratenseElegancy
							.println("Info: Spawning thread 2.");
					stonesoup_thread2.start();
					try {
						Tracer.tracepointMessage("Joining threads");
						Tracer.tracepointMessage("Joining thread-01");
						stonesoup_thread1.join();
						Tracer.tracepointMessage("Joined thread-01");
						Tracer.tracepointMessage("Joining thread-02");
						stonesoup_thread2.join();
						Tracer.tracepointMessage("Joined thread-02");
						Tracer.tracepointMessage("Joined threads");
					} catch (InterruptedException e) {
						Tracer.tracepointError(e.getClass().getName() + ": "
								+ e.getMessage());
						DataInputBlock.ultratenseElegancy
								.println("Interrupted");
					}
					DataInputBlock.ultratenseElegancy
							.println("Info: Threads ended");
				}
			}
			Tracer.tracepointWeaknessEnd();
		}

		private static ReentrantLock lock = new ReentrantLock();

		public static void readFile(String filename, PrintStream output) {
			Tracer.tracepointLocation(
					"/tmp/tmpVvyMIb_ss_testcase/src/src/java/org/apache/poi/poifs/storage/DataInputBlock.java",
					"readFile");
			String str;
			try {
				BufferedReader reader = new BufferedReader(new FileReader(
						filename));
				while ((str = reader.readLine()) != null) {
					output.println(str);
				}
				reader.close();
			} catch (FileNotFoundException e) {
				Tracer.tracepointError("Error reading syncFile.  "
						+ e.getClass().getName() + ": " + e.getMessage());
				output.println("Error reading sync file: " + e);
			} catch (IOException e) {
				Tracer.tracepointError("Error reading syncFile.  "
						+ e.getClass().getName() + ": " + e.getMessage());
				output.println("Error reading sync file: " + e);
			}
		}

		public static class Stonesoup_Str {
			public static StringBuilder data = null;
			public static int size = -1;
		}

		public static void init_Stonesoup_Str(String data, int qsize,
				String filename, PrintStream output) {
			Tracer.tracepointLocation(
					"/tmp/tmpVvyMIb_ss_testcase/src/src/java/org/apache/poi/poifs/storage/DataInputBlock.java",
					"init_Stonesoup_Str");
			output.println(Thread.currentThread().getId()
					+ ": In init_Stonesoup_Str");
			if (Stonesoup_Str.data == null) {
				lock.lock();
				if (Stonesoup_Str.data == null) {
					Tracer.tracepointMessage("CROSSOVER-POINT: BEFORE");
					Stonesoup_Str.data = new StringBuilder();
					Stonesoup_Str.size = data.length();
					output.println(Thread.currentThread().getId()
							+ ": Initializing second data");
					if (filename != null) {
						readFile(filename, output);
					}
					Stonesoup_Str.data.append(data);
					Tracer.tracepointMessage("CROSSOVER-POINT: AFTER");
				} else {
					output.println(Thread.currentThread().getId()
							+ ": No need to initialize");
				}
				lock.unlock();
			} else {
				output.println(Thread.currentThread().getId()
						+ ": Data is already initialized");
			}
		}

		public static class doStuff implements Runnable {
			private int size = 0;
			private String data = null;
			private PrintStream output;
			String filename;

			public void run() {
				Tracer.tracepointLocation(
						"/tmp/tmpVvyMIb_ss_testcase/src/src/java/org/apache/poi/poifs/storage/DataInputBlock.java",
						"doStuff.run");
				try {
					output.println(Thread.currentThread().getId()
							+ ": Inside doStuff");
					Tracer.tracepointMessage("doStuff: entering init_Stonesoup_Str");
					init_Stonesoup_Str(data, size, filename, output);
					output.println(Thread.currentThread().getId()
							+ ": In doStuff Stonesoup_Str is: "
							+ Stonesoup_Str.data.toString());
					Tracer.tracepointVariableString("Stonesoup_Str",
							Stonesoup_Str.data.toString());
				} catch (java.lang.RuntimeException e) {
					e.printStackTrace(output);
					throw e;
				}
			}

			public doStuff(String data, int qsize, String filename,
					PrintStream output) {
				Tracer.tracepointLocation(
						"/tmp/tmpVvyMIb_ss_testcase/src/src/java/org/apache/poi/poifs/storage/DataInputBlock.java",
						"doStuff.ctor");
				this.data = data;
				this.size = qsize;
				this.output = output;
				this.filename = filename;
			}
		}

		public static class doStuff2 implements Runnable {
			private int size = 0;
			private String data = null;
			private PrintStream output;
			private String filename;

			public void run() {
				Tracer.tracepointLocation(
						"/tmp/tmpVvyMIb_ss_testcase/src/src/java/org/apache/poi/poifs/storage/DataInputBlock.java",
						"doStuff2.run");
				int[] sortMe = new int[size];
				try {
					output.println(Thread.currentThread().getId()
							+ ": Inside doStuff2");
					for (int i = 0; i < size; i++) {
						sortMe[i] = size - i;
					}
					Arrays.sort(sortMe);
					readFile(filename, output);
					Tracer.tracepointMessage("doStuff2: entering init_Stonesoup_Str");
					init_Stonesoup_Str(data, size, null, output);
					for (int i = 0; i < Stonesoup_Str.data.length(); i++) {
						if (Stonesoup_Str.data.charAt(i) >= 'a'
								|| Stonesoup_Str.data.charAt(i) <= 'z') {
							Stonesoup_Str.data
									.setCharAt(i, (char) (Stonesoup_Str.data
											.charAt(i) - ('a' - 'A')));
						}
					}
					Tracer.tracepointMessage("TRIGGER-POINT: BEFORE");
					if (Stonesoup_Str.data.charAt(0) != '\0') {
						output.println(Thread.currentThread().getId()
								+ ": In doStuff2 Stonesoup_Str is: "
								+ Stonesoup_Str.data.toString());
					}
					Tracer.tracepointMessage("TRIGGER-POINT: AFTER");
				} catch (java.lang.RuntimeException e) {
					e.printStackTrace(output);
					throw e;
				}
			}

			public doStuff2(String data, int size, String filename,
					PrintStream output) {
				Tracer.tracepointLocation(
						"/tmp/tmpVvyMIb_ss_testcase/src/src/java/org/apache/poi/poifs/storage/DataInputBlock.java",
						"doStuff2.ctor");
				this.data = data;
				this.size = size;
				this.filename = filename;
				this.output = output;
			}
		}
	}
}
